"""ORM client for a specific cell in an inventory in SciNote."""

from abc import ABC, abstractmethod
import logging
from typing import List

from ..client.api.inventory_cell_client import InventoryCellClient
from ..client.models.inventory_cell import (
    Attributes,
    CreateInventoryCell,
    UpdateInventoryCell,
)
from ..client.models.inventory_column_list_item import InventoryColumnListItem

logger = logging.getLogger(__name__)


class Cell(ABC):
    """Base class for a cell in an inventory in SciNote."""

    def __init__(
        self, cell_id: int, attributes: Attributes, client: InventoryCellClient
    ):
        self.cell_id = cell_id
        self.attributes = attributes
        self.__client = client

    def value(self):
        """Get the value of the cell."""
        logger.debug('Getting value for cell %s', self.attributes)

        # Empty cells will not have a value.
        if self.attributes.value is None:
            return None

        return self._do_value()

    @abstractmethod
    def _do_value(self):
        """Get the value of the cell."""
        pass

    async def update_value(self, value: str):
        """Update the value stored in cell."""
        logger.debug(f'Updating value for cell {self.attributes}')
        # We need to check if we are creating a new value or updating an
        # existing one.

        # For some cells like list, we need to convert to the list id from
        # the string that is being used.
        value = self._get_value_for_create_or_update(value)

        if self.attributes.value is None:
            await self.__do_create_cell(value)
        else:
            await self.__do_update_cell(value)

    async def __do_create_cell(self, value: str):
        try:
            result = await self.__client.create_cell(
                CreateInventoryCell(
                    value=value, column_id=self.attributes.column_id
                )
            )
            if result:
                self.cell_id = result.id
                self.attributes = result.attributes
        except Exception as e:
            logger.error(f'Failed to create cell value={value}: {e}')
            raise

    async def __do_update_cell(self, value: str):
        try:
            result = await self.__client.update_cell(
                UpdateInventoryCell(
                    id=self.cell_id,
                    value=value,
                    column_id=self.attributes.column_id,
                )
            )
            if result:
                self.attributes = result.attributes
        except Exception as e:
            logger.error(f'Failed to update cell {self.cell_id}: {e}')
            raise

    @abstractmethod
    def _get_value_for_create_or_update(self, value) -> str:
        """Get the value for create or update."""
        pass

    def match(self, value: str) -> bool:
        """Check if the value of the cell matches the supplied string."""

        if self.attributes.value is None:
            return False

        return self._do_match(value)

    @abstractmethod
    def _do_match(self, value: str) -> bool:
        """Check if the value of the cell matches the supplied string."""
        pass


class ListCell(Cell):
    """ORM client for a list cell in an inventory in SciNote."""

    def __init__(
        self,
        cell_id: int,
        attributes: Attributes,
        client: InventoryCellClient,
        list_column_items: List[InventoryColumnListItem],
    ):
        super().__init__(cell_id, attributes, client)
        self.list_column_items = list_column_items

    def _do_value(self):
        """Get the value of the cell."""
        return self.attributes.value.inventory_list_item_name

    def _do_match(self, value: str) -> bool:
        """Check if the value of the cell matches the supplied string."""
        return value == self.attributes.value.inventory_list_item_name

    def _get_value_for_create_or_update(self, value) -> str:
        """Get the value for create or update."""
        # We need to map the list item data to the list item id
        list_item_id = next(
            (
                item.id
                for item in self.list_column_items
                if item.attributes.data == value
            ),
            None,
        )
        if list_item_id is None:
            raise ValueError(
                f'List item {value} does not exist in column '
                f'{self.attributes.column_id}.'
            )
        return list_item_id


class TextCell(Cell):
    """ORM client for a text cell in an inventory in SciNote."""

    def __init__(
        self, cell_id: int, attributes: Attributes, client: InventoryCellClient
    ):
        super().__init__(cell_id, attributes, client)

    def _do_value(self):
        """Get the value of the cell."""
        return self.attributes.value.text

    def _do_match(self, value: str) -> bool:
        """Check if the value of the cell matches the supplied string."""
        return value == self.attributes.value.text

    def _get_value_for_create_or_update(self, value) -> str:
        """Get the value for create or update."""
        return value


class NumberCell(Cell):
    """ORM client for a number cell in an inventory in SciNote."""

    def __init__(
        self, cell_id: int, attributes: Attributes, client: InventoryCellClient
    ):
        super().__init__(cell_id, attributes, client)

    def _do_value(self):
        """Get the value of the cell."""
        return self.attributes.value.data

    def _do_match(self, value: str) -> bool:
        """Check if the value of the cell matches the supplied string."""
        return value == self.attributes.value.data

    def _get_value_for_create_or_update(self, value) -> str:
        """Get the value for create or update."""
        return value


class DateCell(Cell):
    """ORM client for a date cell in an inventory in SciNote."""

    def __init__(
        self, cell_id: int, attributes: Attributes, client: InventoryCellClient
    ):
        super().__init__(cell_id, attributes, client)

    def _do_value(self):
        """Get the value of the cell."""
        return self.attributes.value.date

    def _do_match(self, value: str) -> bool:
        """Check if the value of the cell matches the supplied string."""
        return value == self.attributes.value.date

    def _get_value_for_create_or_update(self, value) -> str:
        """Get the value for create or update."""
        return value


class DateTimeCell(Cell):
    """ORM client for a specific cell in an inventory in SciNote."""

    def __init__(
        self, cell_id: int, attributes: Attributes, client: InventoryCellClient
    ):
        super().__init__(cell_id, attributes, client)

    def _do_value(self):
        """Get the value of the cell."""
        return self.attributes.value.date_time

    def _do_match(self, value: str) -> bool:
        """Check if the value of the cell matches the supplied string."""
        return value == self.attributes.value.date_time

    def _get_value_for_create_or_update(self, value) -> str:
        """Get the value for create or update."""
        return value


class CellFactory:
    """Factory class for creating cells in an inventory in SciNote."""

    @staticmethod
    def create_cell(
        cell_id: int,
        attributes: Attributes,
        client: InventoryCellClient,
        **kwargs,
    ) -> Cell:
        """Create a new cell in the inventory."""
        match attributes.value_type:
            case 'list':
                if 'column_list_items' not in kwargs:
                    raise ValueError(
                        'Missing column_list_items for list column type.'
                    )
                return ListCell(
                    cell_id, attributes, client, kwargs['column_list_items']
                )
            case 'text':
                return TextCell(cell_id, attributes, client)
            case 'number':
                return NumberCell(cell_id, attributes, client)
            case 'date_time':
                return DateTimeCell(cell_id, attributes, client)
            case 'date':
                return DateCell(cell_id, attributes, client)
            case _:
                raise ValueError(
                    f'Unsupported value type {attributes.value_type}'
                )



"""SciNote rest api client for inventory columns."""

import logging
from typing import List

import httpx
from pydantic import ValidationError

from ..models.inventory_columns import InventoryColumn
from .inventory_column_list_item_client import InventoryColumnListItemClient

logger = logging.getLogger(__name__)


class InventoryColumnClient:
    """Client for interacting with the SciNote REST API."""

    def __init__(
        self, base_url: str, api_key: str, team_id: int, inventory_id: int
    ):
        """Initialize the client."""
        self.base_url = base_url
        self.api_key = api_key
        self.team_id = team_id
        self.inventory_id = inventory_id
        self.headers = {
            'Api-Key': api_key,
        }

    def list_item_client(self, column_id: int) -> InventoryColumnListItemClient:
        """Return a client for interacting with the list items of a column."""
        return InventoryColumnListItemClient(
            self.base_url,
            self.api_key,
            self.team_id,
            self.inventory_id,
            column_id,
        )

    # Inventory column related methods
    async def get_columns(self) -> List[InventoryColumn]:
        """Get all inventory columns from SciNote."""

        # Need to loop through all pages to get all of the columns
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/columns'
        )
        result = []

        # Keep track of visited urls to avoid infinite loops
        visited = set()
        async with httpx.AsyncClient() as client:
            while url:
                response = await client.get(url, headers=self.headers)
                response.raise_for_status()
                visited.add(url)
                logger.debug(f'Get columns response: {response.json()}')
                try:
                    [
                        result.append(InventoryColumn(**column))
                        for column in response.json()['data']
                    ]
                except ValidationError as e:
                    raise ValueError(f'Invalid response from SciNote: {e}')

                url = response.json()['links'].get('next', None)

                # Check we haven't visited this url before
                if url in visited:
                    break

            return result

    async def create_column(
        self, name: str, data_type: str, decimals: int = None
    ) -> str:
        """Create a new inventory column in SciNote."""
        if decimals is not None and data_type not in ['number', 'stock']:
            raise ValueError(
                'Decimals can only be set for columns with data type number'
                ' or stock'
            )

        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/columns'
        )
        json_body = {
            'data': {
                'type': 'inventory_columns',
                'attributes': {
                    'name': name,
                    'data_type': data_type,
                },
            },
        }

        if decimals is not None:
            json_body['data']['attributes']['metadata'] = {'decimals': decimals}

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers=self.headers,
                json=json_body,
            )
            response.raise_for_status()
            return response.json()['data']['id']



"""Model representing an inventory item in SciNote."""

from datetime import datetime

from pydantic import BaseModel, ConfigDict, Field

from .inventory_cell import InventoryCell


class Attributes(BaseModel):
    name: str
    archived: bool = False
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class Relationships(BaseModel):
    inventory_cells: dict


class InventoryItem(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    type: str
    attributes: Attributes
    relationships: Relationships

    # This is not present in the schema from SciNote, we add it here so
    # we can retrieve the cells while we're retrieving the items.
    inventory_cells: list[InventoryCell] = []


class UpdateIncludedCellsAttributes(BaseModel):
    value: str | int
    column_id: int


class UpdateIncludedCells(BaseModel):
    model_config = ConfigDict(extra='ignore')
    id: str
    type: str = 'inventory_cells'
    attributes: UpdateIncludedCellsAttributes


class UpdateItem(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    type: str = 'inventory_items'
    attributes: Attributes
    included: list[UpdateIncludedCells] = []



"""Model representing an inventory cell in SciNote."""

from datetime import date, datetime
from typing import List

from pydantic import BaseModel, ConfigDict, Field


class TextItem(BaseModel):
    text: str


class ListItem(BaseModel):
    inventory_list_item_id: int
    inventory_list_item_name: str
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class CheckListItem(BaseModel):
    inventory_checklist_item_ids: List[int]
    inventory_checklist_item_names: List[str]
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class StatusItem(BaseModel):
    inventory_status_item_id: int
    inventory_status_item_name: str
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class FileItem(BaseModel):
    file_id: int
    file_name: str
    file_size: int
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class DateTimeItem(BaseModel):
    date_time: datetime
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class DateItem(BaseModel):
    date: date
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class NumberItem(BaseModel):
    data: str
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class Attributes(BaseModel):
    value_type: str
    value: (
        TextItem
        | ListItem
        | ListItem
        | CheckListItem
        | StatusItem
        | FileItem
        | DateTimeItem
        | DateItem
        | NumberItem
    ) = None
    column_id: int
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class InventoryCell(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    type: str = 'inventory_cells'
    attributes: Attributes


class CreateInventoryCell(BaseModel):
    """Model represents a new inventory cell in SciNote."""

    type: str = 'inventory_cells'
    value: str | int
    column_id: int


class UpdateInventoryCell(BaseModel):
    """Model represents updating an inventory cell in SciNote."""

    id: str
    type: str = 'inventory_cells'
    value: str | int
    column_id: int



"""Maps a SciNote item to an ORM item."""

from datetime import datetime
import logging
from typing import List

from ..client.api.inventory_item_client import InventoryItemClient
from ..client.models.inventory_cell import Attributes, InventoryCell
from ..client.models.inventory_column_list_item import InventoryColumnListItem
from ..client.models.inventory_item import UpdateItem
from .cell import Cell, CellFactory

logger = logging.getLogger(__name__)
"""
An item is a complete row in the inventory.

When SciNote returns an item, the values are mapped to a column ID. We need to
do this mapping here using the columns for the inventory and the values that
are passed into the constructor.
"""


class Item:
    """Maps a SciNote item to an ORM item."""

    def __init__(
        self,
        id: str,
        name: str,
        created_at: datetime,
        item_client: InventoryItemClient,
        columns: dict,
        column_list_items: dict[str, List[InventoryColumnListItem]],
        values: list[InventoryCell],
    ):
        self.id = id
        self.name = name
        self.created_at = created_at
        self.item_client = item_client
        self.column_ids = {}
        self.column_names = {}
        self.column_list_items = column_list_items
        self.__cells = {}

        logger.debug('Columns: %s', columns)
        logger.debug('Values: %s', values)

        # Create a map of column ids to names and vice versa.
        for index, (key, value) in enumerate(columns.items()):
            name = key.lower().replace(' ', '_')
            self.column_ids[value.id] = name
            self.column_names[name] = value.id

        # Create cells for those cells that have values in SciNote.
        self.__create_cells_from_inventory_cells(values)

        # SciNote will only return values if they are not empty. We need to
        # create empty cells for the columns that are not present.
        empty_columns = set(self.column_names.keys()) - set(self.__cells.keys())
        logger.debug('Empty columns: %s', empty_columns)
        for column_name in empty_columns:
            # Get metadata for the missing column.
            column = columns[column_name]
            list_items = column_list_items.get(column.id, None)
            self.__cells[column_name] = CellFactory.create_cell(
                None,  # No id for cells that do not exist.
                Attributes(
                    column_id=column.id,
                    value_type=column.attributes.data_type,
                ),
                item_client.cell_client(self.id),
                column_list_items=list_items,
            )

    def cell(self, name) -> Cell:
        """Get the cell by name."""
        if self.column_names.get(name) is None:
            raise ValueError(f'Item {name} not found')
        return self.__cells.get(name)

    def cells(self) -> list[Cell]:
        """Get all of the cells."""
        return [value for value in self.__cells.values()]

    def match(self, **kwargs) -> bool:
        """Match the item against the given values."""
        for key, value in kwargs.items():
            if key == 'name':
                if self.name != value:
                    return False
                continue
            if key == 'id':
                if self.id != value:
                    return False
                continue

            if key not in self.column_names:
                logger.warning('Column %s not found in item', key)
                return False

            if not self.cell(key).match(value):
                return False

        return True

    async def update(self, **kwargs) -> 'Item':
        """Update the item and its cells with the given values.

        This allows for a bulk update of the item and its cells in one call to
        SciNote. The caller passes keyword arguments for the cells to update,
        where the key is the name of the column and the value is the value to
        update that cell to.

        The caller can also pass the key 'name', which represents the name of
        the item.

        The method returns the updated Item object.
        """
        # only update the item name if the call to SciNote succeeds.
        item_name = self.name
        included = []
        for key, value in kwargs.items():
            if key == 'name':
                item_name = value
                continue

            if key not in self.column_names:
                logger.error('Column %s not found in item', key)
                raise KeyError(
                    f"Column '{key}' not found in item. Update aborted."
                )

            cell = self.__cells.get(key)

            if cell is None:
                raise KeyError(
                    f"Cell '{key}' not found in item. Update aborted."
                )

            # For the update we need the column id and the value to update.
            included.append(
                {
                    'id': cell.cell_id,
                    'attributes': {
                        'value': cell._get_value_for_create_or_update(value),
                        'column_id': cell.attributes.column_id,
                    },
                }
            )

        update = UpdateItem(
            id=self.id,
            attributes={
                'name': item_name,
            },
            included=included,
        )

        result = await self.item_client.update_item(update)
        self.name = result.attributes.name
        # Update the cells for each column that was updated. We just overwrite
        # any existing cells.
        self.__create_cells_from_inventory_cells(result.inventory_cells)

        return self

    def __create_cells_from_inventory_cells(
        self, values: List[InventoryCell]
    ) -> None:
        """Create cells from a list of inventory cells."""

        # For each incoming value, map it to the column name. There is added
        # complexity as the name of the field that is the value of the cell
        # is dependent on the type of the column.
        for value in values:
            column_id = str(value.attributes.column_id)
            column_name = self.column_ids[column_id]
            list_items = self.column_list_items.get(column_id, None)

            self.__cells[column_name] = CellFactory.create_cell(
                value.id,
                value.attributes,
                self.item_client.cell_client(self.id),
                column_list_items=list_items,
            )



"""SciNote REST API client for inventory items."""

import logging
from typing import List

import httpx
from pydantic import ValidationError

from ..models.inventory_cell import CreateInventoryCell, InventoryCell
from ..models.inventory_item import InventoryItem, UpdateItem
from .inventory_cell_client import InventoryCellClient
from .inventory_item_child_client import InventoryChildClient
from .inventory_item_parent_client import InventoryParentClient

logger = logging.getLogger(__name__)


class InventoryItemClient:
    """Client for interacting with the SciNote REST API."""

    def __init__(
        self, base_url: str, api_key: str, team_id: int, inventory_id: int
    ):
        """Initialize the client."""
        self.base_url = base_url
        self.api_key = api_key
        self.team_id = team_id
        self.inventory_id = inventory_id
        self.headers = {
            'Api-Key': api_key,
        }
        self.params = {
            'include': 'inventory_cells',
        }

    def cell_client(self, item: int) -> InventoryCellClient:
        return InventoryCellClient(
            self.base_url, self.api_key, self.team_id, self.inventory_id, item
        )

    def parent_client(self, item: int) -> InventoryParentClient:
        return InventoryParentClient(
            self.base_url, self.api_key, self.team_id, self.inventory_id, item
        )

    def child_client(self, item: int) -> InventoryChildClient:
        return InventoryChildClient(
            self.base_url, self.api_key, self.team_id, self.inventory_id, item
        )

    async def get_items(self) -> List[InventoryItem]:
        """Get all inventory items from SciNote."""

        # Need to loop through all pages to get all of the items
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items'
        )
        result = []
        inventory_cells = {}

        # Keep track of visited urls to avoid infinite loops
        visited = set()
        async with httpx.AsyncClient() as client:
            while url:
                response = await client.get(
                    url, headers=self.headers, params=self.params
                )
                response.raise_for_status()
                visited.add(url)
                try:
                    [
                        result.append(InventoryItem(**item))
                        for item in response.json()['data']
                    ]
                except ValidationError as e:
                    raise ValueError(f'Invalid response from SciNote: {e}')

                # Read of all the cell responses as well so we can marry them
                # together later.
                try:
                    [
                        inventory_cells.update(
                            {cell['id']: InventoryCell(**cell)}
                        )
                        for cell in response.json()['included']
                    ]
                except ValidationError as e:
                    raise ValueError(f'Invalid response from SciNote: {e}')

                url = response.json()['links'].get('next', None)

                # Check we haven't visited this url before
                if url in visited:
                    break

            # If we have inventory cells, we need to attach them to the items.
            for item in result:
                item.inventory_cells = [
                    inventory_cells[cell['id']]
                    for cell in item.relationships.inventory_cells['data']
                ]

            return result

    async def create_item(
        self, name: str, cells: List[CreateInventoryCell] = None
    ) -> InventoryItem:
        """Create a new inventory item in SciNote."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items'
        )

        included = []
        if cells:
            for cell in cells:
                included.append(
                    {
                        'type': cell.type,
                        'attributes': {
                            'value': cell.value,
                            'column_id': cell.column_id,
                        },
                    }
                )

        json_body = {
            'data': {
                'type': 'inventory_items',
                'attributes': {
                    'name': name,
                },
            },
            'included': included,
        }

        async with httpx.AsyncClient() as client:
            request = client.build_request(
                'POST',
                url,
                headers=self.headers,
                json=json_body,
            )

            logger.debug(
                'Creating item with request: %s', request.content.decode()
            )

            response = await client.send(request)
            response.raise_for_status()
            item = InventoryItem(**response.json()['data'])

            if 'included' in response.json():
                for cell in response.json()['included']:
                    item.inventory_cells.append(InventoryCell(**cell))

            return item

    async def update_item(self, update: UpdateItem) -> InventoryItem:
        """Update an existing inventory item in SciNote."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items/{update.id}'
        )

        included = []
        if update.included:
            for cell in update.included:
                included.append(
                    {
                        'id': cell.id,
                        'type': cell.type,
                        'attributes': {
                            'value': cell.attributes.value,
                            'column_id': cell.attributes.column_id,
                        },
                    }
                )

        json_body = {
            'data': {
                'id': update.id,
                'type': update.type,
                'attributes': {
                    'name': update.attributes.name,
                },
            },
            'included': included,
        }

        async with httpx.AsyncClient() as client:
            logger.debug('url: %s', url)
            request = client.build_request(
                'PATCH',
                url,
                headers=self.headers,
                json=json_body,
            )

            logger.debug(
                'Updating item with request: %s', request.content.decode()
            )

            response = await client.send(request)
            response.raise_for_status()
            item = InventoryItem(**response.json()['data'])

            if 'included' in response.json():
                for cell in response.json()['included']:
                    item.inventory_cells.append(InventoryCell(**cell))

            return item
